import copy
import fnmatch
import inspect
from typing import Any
from typing import Dict
from typing import List


def gen_tag(hub, chunk):
    """
    Generate the unique tag used to track the execution of the chunk
    """
    return f'{chunk["state"]}_|-{chunk["__id__"]}_|-{chunk["name"]}_|-{chunk["fun"]}'


def get_chunks(hub, low, state, name):
    """
    Search in the low state for the chunk with the given designation
    """
    rets = []
    for chunk in low:
        if state == "sls":
            if fnmatch.fnmatch(chunk["__sls__"], name):
                rets.append(chunk)
                continue
        if state == chunk["state"]:
            if fnmatch.fnmatch(chunk["name"], name) or fnmatch.fnmatch(
                chunk["__id__"], name
            ):
                rets.append(chunk)
    return rets


def find_name(hub, name, state, high):
    """
    Scan high data for the id referencing the given name and return a list of (IDs, state) tuples that match
    Note: if `state` is sls, then we are looking for all IDs that match the given SLS
    """
    ext_id = []
    if name in high:
        ext_id.append((name, state))
    # if we are requiring an entire SLS, then we need to add ourselves to everything in that SLS
    elif state == "sls":
        for nid, item in high.items():
            if item["__sls__"] == name:
                ext_id.append((nid, next(iter(item))))
    # otherwise we are requiring a single state, lets find it
    else:
        # We need to scan for the name
        for nid in high:
            if state in high[nid]:
                if isinstance(high[nid][state], list):
                    for arg in high[nid][state]:
                        if not isinstance(arg, dict):
                            continue
                        if len(arg) != 1:
                            continue
                        if arg[next(iter(arg))] == name:
                            ext_id.append((nid, state))
    return ext_id


def format_call(
    hub,
    fun,
    data,
    ignore_changes: List = None,
    initial_ret=None,
    expected_extra_kws=(),
    is_class_method=None,
    enforced_state: Dict[str, Any] = None,
):
    """
    Build the required arguments and keyword arguments required for the passed
    function.
    :param fun: The function to get the argspec from
    :param data: A dictionary containing the required data to build the
                 arguments and keyword arguments.
    :param ignore_changes: A list of path of parameters that will be assigned to None value
                           to ignore being updated in present().
    :param initial_ret: The initial return data pre-populated as dictionary or
                        None
    :param expected_extra_kws: Any expected extra keyword argument names which
                               should not trigger a :ref:`SaltInvocationError`
    :param is_class_method: Pass True if you are sure that the function being passed
                            is a class method. The reason for this is that on Python 3
                            ``inspect.ismethod`` only returns ``True`` for bound methods,
                            while on Python 2, it returns ``True`` for bound and unbound
                            methods. So, on Python 3, in case of a class method, you'd
                            need the class to which the function belongs to be instantiated
                            and this is not always wanted.
    :param enforced_state: A dictionary with the parameters from a previous run
    :returns: A dictionary with the function required arguments and keyword
              arguments.
    """
    ret = initial_ret is not None and initial_ret or {}
    enforced_state = enforced_state or {}

    ret["args"] = []
    ret["kwargs"] = {}
    args = []
    kwargs = {}
    keywords = False

    sig = fun.signature
    for name, param in sig.parameters.items():
        if name == "hub":
            continue
        elif param.kind.name == "POSITIONAL_OR_KEYWORD":
            if isinstance(param.default, inspect._empty):
                args.append(name)
            else:
                # get it from enforced state before default
                kwargs[name] = enforced_state.get(name, param.default)
        elif param.kind.name == "KEYWORD_ONLY":
            # get it from enforced state before default
            kwargs[name] = enforced_state.get(name, param.default)
        elif param.kind.name == "VAR_KEYWORD":
            keywords = True
    ret["avail_kwargs"] = copy.copy(kwargs)
    ret["avail_args"] = copy.copy(args)
    ret["keywords"] = keywords

    # Since we WILL be changing the data dictionary, let's change a copy of it
    data = data.copy()
    is_existing_resource = bool(enforced_state) or bool(data.get("resource_id"))

    missing_args = []

    for key in kwargs:
        try:
            # Do not override enforced_state value with None,
            # unless no value is set
            val = data.pop(key)
            if val is not None or kwargs[key] == inspect._empty:
                kwargs[key] = val
        except KeyError:
            # Let's leave the default value in place
            pass

    # If ignore_changes contains parameters, then we try to assign those parameters with None to skip updating.
    if ignore_changes and is_existing_resource:
        hub.idem.tool.ignore_changes.ignore_parameter_changes(
            ignore_changes=ignore_changes,
            params=kwargs,
            param_signatures=sig.parameters,
        )

    while args:
        arg = args.pop(0)
        try:
            ret["args"].append(data.pop(arg))
        except KeyError:
            missing_args.append(arg)

    if missing_args:
        used_args_count = len(ret["args"]) + len(args)
        args_count = used_args_count + len(missing_args)
        raise RuntimeError(
            f"{fun.name} takes at least {args_count} arguments ({used_args_count} given)"
        )

    ret["kwargs"].update(kwargs)

    if keywords:
        # The function accepts **kwargs, any non expected extra keyword
        # arguments will made available.
        for key, value in data.items():
            if key in expected_extra_kws:
                continue
            ret["kwargs"][key] = value

        # No need to check for extra keyword arguments since they are all
        # **kwargs now. Return
        return ret

    # Did not return yet? Lets gather any remaining and unexpected keyword
    # arguments
    extra = {}
    for key, value in data.items():
        if key in expected_extra_kws:
            continue
        extra[key] = copy.deepcopy(value)

    if extra:
        # Found unexpected keyword arguments, raise an error to the user
        if len(extra) == 1:
            msg = "'{0[0]}' is an invalid keyword argument for '{1}'".format(
                list(extra.keys()),
                ret.get(
                    # In case this is being called for a state module
                    "full",
                    # Not a state module, build the name
                    f"{fun.__module__}.{fun.__name__}",
                ),
            )
        else:
            msg = "{} and '{}' are invalid keyword arguments for '{}'".format(
                ", ".join([f"'{e}'" for e in extra][:-1]),
                list(extra.keys())[-1],
                ret.get(
                    # In case this is being called for a state module
                    "full",
                    # Not a state module, build the name
                    f"{fun.__module__}.{fun.__name__}",
                ),
            )
        # raise SaltInvocationError(msg)
    return ret


def ishashable(hub, obj):
    """
    A simple test to verify if a given object is hashable and can therefore
    be used as a key in a dict
    """
    try:
        hash(obj)
    except TypeError:
        return False
    return True


def iter_high(hub, high):
    """
    Take a high-state structure and iterate over it yielding the elements down to the
    execution args
    Yields (id_, body, state, run, arg)
    """
    for id_, body in high.items():
        if not isinstance(body, dict):
            continue
        for state, run in body.items():
            if state.startswith("__"):
                continue
            for arg in run:
                yield id_, body, state, run, arg


def iter_high_leaf_args(hub, high):
    """
    Take a high-state structure and iterate over it yielding the elements down to the
    execution args
    Yields (id_, body, state, run, arg)
    """
    for id_, body in high.items():
        if not isinstance(body, dict):
            continue
        for state, run in body.items():
            if state.startswith("__"):
                continue
            for arg in run:
                yield from _iter_arg_chain(id_, body, state, run, "", arg)


def _iter_arg_chain(id_, body, state, run, arg_key_def, arg):
    if isinstance(arg, dict):
        for arg_key, arg_value in arg.items():
            # Escape dictionary references in the key definition to skip evaluating by arg_bind resolver.
            arg_key = arg_key.replace("[", "[\\")
            if arg_key_def:
                current_arg_key_def = arg_key_def + ":" + arg_key
            else:
                current_arg_key_def = arg_key

            if isinstance(arg_value, dict):
                yield from _iter_arg_chain(
                    id_, body, state, run, current_arg_key_def, arg_value
                )
            elif isinstance(arg_value, list):
                for index, arg_value_item in enumerate(arg_value):
                    yield from _iter_arg_chain(
                        id_,
                        body,
                        state,
                        run,
                        current_arg_key_def + f"[{index}]",
                        arg_value_item,
                    )
            else:
                yield id_, body, state, run, current_arg_key_def, arg_value
    elif isinstance(arg, list):
        for index, arg_value_item in enumerate(arg):
            yield from _iter_arg_chain(
                id_, body, state, run, arg_key_def + f"[{index}]", arg_value_item
            )
    else:
        yield id_, body, state, run, arg_key_def, arg



from typing import Any
from typing import Dict


def run(
    hub,
    seq: Dict[int, Dict[str, Any]],
    low: Dict[str, Any],
    running: Dict[str, Any],
    options: Dict[str, Any],
) -> Dict[int, Dict[str, Any]]:
    """
    Process the requisites from ESM
    This function will check if the requisites for argument binding is present in this run
    If its present we do not do anything as requisites from the current run will already be added.
    If the requisite is not present in this run we check the ESM
    if the resource is already run previously, and it exists in ESM,
    we prepare the requisite and add to the resource requisites.

    This function will only add requisites to resources if they are not present in current run and exists in ESM
    This will not modify the current running resources nor this will send any events
    We only read a state from ESM to add to requisites. The state is not written back to ESM nor added to current run.
    """
    for ind, data in seq.items():
        if "arg_bind" not in data["chunk"]:
            continue
        chunk = data["chunk"]
        for rdef in chunk["arg_bind"]:
            if not isinstance(rdef, dict):
                data["errors"].append(f"{rdef} should be dictionary")
                continue
            state = next(iter(rdef))
            if isinstance(rdef[state], list):
                name_defs = rdef[state]
            else:
                name_defs = [{rdef[state]: []}]

            for name_def in name_defs:
                if not isinstance(name_def, dict):
                    data["errors"].append(f"{name_def} should be dictionary")
                    continue
                name = next(iter(name_def))
                args = name_def[name]
                r_chunks = hub.idem.tools.get_chunks(low, state, name)
                if not r_chunks:
                    # If r_chunks is not found in current run check if its present in ESM
                    hub.log.debug(
                        f"Requisite arg_bind {state}:{name} not found in current run. checking in ESM for arg_bind requisite"
                    )
                    r_chunks_from_esm = hub.tool.idem.esm.get_chunks_from_esm(
                        state, name
                    )
                    if not r_chunks_from_esm:
                        data["errors"].append(
                            f"Requisite arg_bind {state}:{name} not found"
                        )
                    for r_chunk in r_chunks_from_esm:
                        reqret_esm = hub.tool.idem.esm.update_running_from_esm(r_chunk)
                        r_tag = hub.idem.tools.gen_tag(r_chunk)
                        reqret = {
                            "req": "arg_bind",
                            "name": name,
                            "state": state,
                            "r_tag": r_tag,
                            "ret": reqret_esm,
                            "chunk": "r_chunk",
                            "args": args,
                        }
                        data["reqrets"].append(reqret)
    return seq



from typing import Dict
from typing import List

import dict_tools.data as data


def __init__(hub):
    # Keep track of the number of times the idem runtime is called for the progress bar
    hub.idem.run.init.RUNTIME_ITERATIONS = 0


async def start(hub, name, pending_tags: List = None, target: str = None):
    """
    Called only after the named run has compiled low data. If no low data
    is present an exception will be raised
    pending_tags: list of state tags to re-run, passed from reconciliation.
                  If None all states are executed.
    """
    if not hub.idem.RUNS[name].get("low"):
        raise ValueError(f"No low data for '{name}' in RUNS")
    ctx = data.NamespaceDict({"run_name": name, "test": hub.idem.RUNS[name]["test"]})

    rtime = hub.idem.RUNS[name]["runtime"]
    low = hub.idem.RUNS[name].get("low")
    # Add metadata to second run

    # Fire an event with all the pre-rendered low-data for all states
    await hub.idem.event.put(
        profile="idem-low",
        body=low,
        tags={"ref": "idem.run.init.start", "type": "state-low-data"},
    )
    old_seq = {}
    old_seq_len = -1
    needs_post_low = True
    options = {"invert_state": hub.idem.RUNS[name]["invert_state"]}
    low_declaration_ids = [item.get("__id__") for item in low]

    # If there are pending tags re-run only those (reconciliation)
    if pending_tags:
        # Clear previous running data
        hub.idem.RUNS[name]["running"] = {}
        # Keep low data only for pending tags
        declaration_ids = [tag.split("_|-")[1] for tag in pending_tags]
        low = [item for item in low if item.get("__id__") in declaration_ids]

    if target:
        # If target is specified - identify all its dependencies
        low_items = gather_target_and_reqs(low, target)
        if len(low_items) < 1:
            raise ValueError(
                f"Invalid 'target' for run '{name}': {target}. 'target' should be a declaration ID."
            )

        declaration_ids = [item.get("__id__") for item in low_items]
        hub.idem.RUNS[name]["high"] = {
            key: item
            for (key, item) in hub.idem.RUNS[name]["high"].items()
            if key in declaration_ids
        }
        low = low_items

    while True:
        seq = hub.idem.req.seq.init.run(
            None, low, hub.idem.RUNS[name]["running"], options
        )
        if seq == old_seq:
            unmet_reqs = {chunk["tag"]: chunk["unmet"] for chunk in seq.values()}
            if len(unmet_reqs) == 0:
                raise Exception(f"Invalid syntax for '{name}'")
            raise Exception(f"No sequence changed for '{name}': {unmet_reqs}")

        # Initialize the progress bar
        progress_bar = hub.tool.progress.init.create(
            seq,
            desc=f"idem runtime: {hub.idem.run.init.RUNTIME_ITERATIONS}",
            unit="states",
        )
        # Increment the runtime iteration number after using it for the progress bar
        hub.idem.run.init.RUNTIME_ITERATIONS += 1

        await hub.idem.run[rtime].runtime(
            name,
            ctx,
            seq,
            low,
            hub.idem.RUNS[name]["running"],
            hub.idem.RUNS[name]["managed_state"],
            progress=progress_bar,
        )

        render_data = await hub.idem.resolve.init.render(
            name,
            blocks=hub.idem.RUNS[name]["blocks"],
            sls_refs=hub.idem.RUNS[name]["sls_refs"],
            resolved=hub.idem.RUNS[name]["resolved"],
        )
        await hub.idem.sls_source.init.update(name, render_data)

        await hub.idem.state.compile(name)

        extend_low(hub, name, low, low_declaration_ids)

        await hub.idem.event.put(
            profile="idem-low",
            body=low,
            tags={"ref": "idem.run.init.start", "type": "state-low-data"},
        )
        if len(low) <= len(hub.idem.RUNS[name]["running"]):
            if hub.idem.RUNS[name]["post_low"] and needs_post_low:
                hub.idem.RUNS[name]["low"].extend(hub.idem.RUNS[name]["post_low"])
                needs_post_low = False
                extend_low(hub, name, low, low_declaration_ids)
                continue
            else:
                break
        if len(seq) == old_seq_len:
            raise RecursionError(f"No progress made on '{name}', Recursive Requisite!")
        old_seq = seq
        old_seq_len = len(seq)


def gather_target_and_reqs(low, target) -> List[dict]:
    """
    Gather the target state(s) and all its requisites
    :param low: low data
    :param target: target name: should match declaration id
    :return: list of chunks
    """
    low_items = []

    # Target must be declaration id
    # It might resolve to multiple target items if there are
    # multiple resources under the same declaration id
    gather_(low, [target], low_items)

    return low_items


def gather_(low: List[Dict], reqs: List[str], low_items: List[Dict]):
    # Recursively gather low items of target and pre-requisites
    # Target or any requisite might resolve to multiple resource
    # items if there are multiple resources under the same declaration id
    if not reqs:
        return
    for req in reqs:
        req_items = [item for item in low if item.get("__id__") == req]
        if len(req_items) < 1:
            continue
        low_items.extend(req_items)
        for req_item in req_items:
            if req_item.get("require"):
                for required in req_item.get("require"):
                    gather_(low, list(required.values()), low_items)


def extend_low(hub, name, low, low_declaration_ids):
    # If new chunks were added to "low" (post_low etc)
    # Add them in case low was modified by 'pending_tags' or 'target'
    new_low_items = [
        item
        for item in hub.idem.RUNS[name].get("low")
        if item.get("__id__") not in low_declaration_ids
    ]
    low.extend(new_low_items)
    low_declaration_ids += [item.get("__id__") for item in new_low_items]



def is_pending(hub, ret: dict, state: str = None) -> bool:
    """
    Default implementation of pending plugin
    Returns 'True' when the state is still 'pending' and reconciliation is required.

    If the state implements is_pending() call the state's implementation,
    otherwise state is pending until 'result' is 'True' and there are no 'changes'.

    :param hub: The hub
    :param ret: (dict) Returned structure of a run
    :param state: (Text, Optional) The name of the state
    :return: bool
    """

    if (
        state is not None
        and hub.states[state] is not None
        and callable(getattr(hub.states[state], "is_pending", None))
    ):
        return hub.states[state].is_pending(ret=ret)

    return not ret["result"] is True or bool(ret["changes"])



from typing import Any
from typing import Dict

from dict_tools.data import SafeNamespaceDict


async def modify(hub, name: str, chunk: Dict[str, Any]) -> Dict[str, Any]:
    """
    Check the state containing the target func and call the mod_creds
    function if present. Therefore gathering the list of creds systems
    to use
    """
    run_name = chunk["ctx"]["run_name"]
    state = chunk["state"].split(".")[0]
    subs = []
    if hasattr(hub, f"states.{state}.ACCT"):
        subs = getattr(hub, f"states.{state}.ACCT")
    elif hasattr(hub, f"{state}.ACCT"):
        subs = getattr(hub, f"{state}.ACCT")
    elif hasattr(hub, f"exec.{state}.ACCT"):
        subs = getattr(hub, f"exec.{state}.ACCT")
    elif hasattr(hub, f"tool.{state}.ACCT"):
        subs = getattr(hub, f"tool.{state}.ACCT")
    hub.log.debug(f"Loaded acct from subs: {subs}")

    # Get the acct_profile that is being used from the chunk and fallback to the acct_profile in the RUN
    profile = chunk.pop("acct_profile", hub.idem.RUNS[run_name]["acct_profile"])

    # Some idem tests or projects that call idem from pure python specify a sub other than "states" for state modules
    # Check for state modules being in alternate locations but default to "states"
    sub = "states"
    for sub in hub.idem.RUNS[run_name]["subs"]:
        try:
            # Get the keyword arguments of the state function that will be called
            function_parameters = hub[sub][chunk["state"]][
                chunk["fun"]
            ].signature.parameters
            break
        except AttributeError:
            ...
    else:
        function_parameters = {}

    # If the acct_profile was in the function parameters, then add back the acct_profile that is being used
    if "acct_profile" in function_parameters:
        chunk["acct_profile"] = profile

    # If acct_data was in the function parameters and in the chunk, then return the chunk without constructing ctx.acct
    # The function is going to decide what to do with the acct_data
    if "acct_data" in function_parameters and chunk.get("acct_data"):
        # Allow acct_data in the function parameters, but protect it from being exposed in events and logs
        chunk["acct_data"] = SafeNamespaceDict(chunk["acct_data"])
        return chunk

    # Retrieve the acct_data from the state chunk and fallback to the acct_data from the RUN
    acct_data = chunk.pop("acct_data", hub.idem.RUNS[run_name]["acct_data"])

    # Validate that the specified acct_profile is specified is within the profiles
    try:
        if hub[sub][state] and hub[sub][state].ACCT:
            if (
                profile
                and acct_data
                and acct_data["profiles"].get(state)
                and profile not in acct_data["profiles"][state]
            ):
                msg = f"Could not find profile '{profile}' in profiles: {acct_data['profiles'][state].keys()}"
                hub.log.error(msg)
                raise Exception(msg)
    # hub.states[state].ACCT may not exists
    except AttributeError:
        ...

    hub.log.debug(f"Loaded profile: {profile}")

    chunk["ctx"]["acct"] = await hub.acct.init.gather(
        subs,
        profile,
        profiles=acct_data.get("profiles"),
        hard_fail=True,
    )

    # Get metadata for the profile
    chunk["ctx"]["acct_details"] = await hub.acct.metadata.init.gather(
        profile=chunk["ctx"]["acct"],
        providers=subs,
    )

    return chunk



from typing import Any
from typing import Dict

__func_alias__ = {"ctx_": "ctx"}
__contracts__ = ["returns", "soft_fail"]


def __init__(hub):
    hub.exec.test.ITEMS = {}
    hub.exec.test.ACCT = ["test"]


def ping(hub):
    """
    Immediately return success
    """
    return {"result": True, "ret": True}


async def aping(hub):
    return {"result": True, "ret": True}


def ctx_(hub, ctx):
    return {"result": True, "comment": None, "ret": ctx}


def fail(hub):
    raise Exception("Expected failure")


async def afail(hub):
    raise Exception("Expected failure")


def echo(hub, value):
    """
    Return the parameter passed in without changing it at all
    """
    return value


async def aecho(hub, value):
    """
    Return the parameter passed in without changing it at all
    """
    return value


def more(hub, ctx, *args, **kwargs):
    """
    Return the ctx and all parameters passed to this state.
    """
    return {
        "result": True,
        "comment": None,
        "ret": {"args": args, "kwargs": kwargs, "ctx": ctx},
    }


async def event(
    hub,
    body: str,
    ingress_profile: str = "default",
    tags: Dict[str, Any] = None,
):
    """
    :param body: The event body
    :param ingress_profile: The ingress profile to allowlist for this event

    .. code-block:: bash

        $ idem exec test.event body ingress_profile="default" --serialize-plugin="json"
        $ idem exec test.event body="my_event" ingress_profile="default" --serialize-plugin="json"
    """
    event_kwargs = dict(body=body, profile=ingress_profile, tags=tags)
    await hub.idem.event.put(**event_kwargs)
    return {"result": True, "ret": event_kwargs}



"""
This file contains routines to get sls files from references
"""
import pathlib
import re
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

SOURCE_PATTERN = re.compile(
    r"^(?P<protocol_plugin>(?P<protocol>\w+)[\+\w+]*)://(?:(?P<profile>[-.\w]+)@)?(?P<data>.+)$"
)


def parse_source(hub, source: str) -> Tuple[str, str, str, str]:
    match = SOURCE_PATTERN.match(source)
    if not match:
        raise ValueError(f"SLS source is not a valid pattern: {source}")
    protocol = match.group("protocol")
    protocol_plugin = match.group("protocol_plugin")
    acct_profile = match.group("profile")
    data = match.group("data")
    return protocol, protocol_plugin, acct_profile, data


async def ref(hub, name: str, sls: str, sources: List[str]) -> Dict[str, Any]:
    """
    Cache the given file from the named reference point

    :param hub:
    :param name: The state run name
    :param sls: an SLS location within the given sources
    :param sources: sls-sources or params-sources

    :returns A file_name/identifier and encoded yaml content

    If the sls sources uses `acct` for authentication, then a profile is specified as part of the sls_source

    .. code-block::

        [proto]://[acct_profile]@[path]
    """
    acct_data = hub.idem.RUNS[name]["acct_data"]
    original_sls = sls

    # Search for the sls ref in each source (in the order the sources were defined) and return the first match
    for source in sources:
        protocol, protocol_plugin, acct_profile, data = hub.idem.get.parse_source(
            source
        )
        if not acct_profile:
            acct_profile = "default"

        # Generate a ctx with relevant credentials for the "source" plugin
        ctx = await hub.idem.acct.ctx(
            protocol_plugin,
            acct_profile=acct_profile,
            acct_data=acct_data,
        )

        # Find the total possible number of parent directories in this source
        possible_parents = len(source.split("./\\"))
        # After each failure, imply a deeper level of relative parent

        for _ in range(possible_parents):
            # Try both posix and windows paths as the sls may be not be on the local filesystem
            posix_path = pathlib.PurePosixPath(sls.replace(".", "/"))
            nt_path = pathlib.PureWindowsPath(sls.replace(".", "\\"))
            sls = str(posix_path).replace("/", ".")

            locations = [
                # Append ".sls" to the tried paths
                (str(posix_path.with_suffix(".sls")), sls),
                (str(nt_path.with_suffix(".sls")), sls),
                # Append "init.sls" to the tried paths
                (str(posix_path.joinpath("init.sls")), f"{sls}.init"),
                (str(nt_path.joinpath("init.sls")), f"{sls}.init"),
                # Try the raw paths
                (str(posix_path), sls),
                (str(nt_path), sls),
                # try the unaltered sls ref
                (sls, sls),
                (f"{sls}.init", f"{sls}.init"),
            ]

            # Try each possible location until we resolve the sls ref
            for location, actual_ref in locations:
                hub.log.debug(f"Trying to find ref '{actual_ref}' in '{location}'")
                processed_location, cache_data = await hub.source[
                    protocol_plugin
                ].cache(ctx, protocol=protocol, source=data, location=location)
                if cache_data:
                    # The reference was found and contained valid data
                    hub.log.debug(f"Found ref '{actual_ref}' in '{processed_location}'")
                    return {
                        "location": processed_location,
                        "sls": actual_ref,
                        "content": cache_data,
                    }
        # Try again with a relative source
        sls = f".{sls}"
    else:
        raise LookupError(f"Could not find SLS ref '{original_sls}' in sources")
